# This code is written at BigVision LLC. It is based on the OpenCV project. It is subject to the license terms in the LICENSE file found in this distribution and at http://opencv.org/license.html

# Usage example:  python3 object_detection_yolo.py --video=run.mp4
#                 python3 object_detection_yolo.py --image=bird.jpg

import argparse
import os.path
import sys

import numpy as np

import cv2

# Initialize the parameters
confThreshold = 0.5  # Confidence threshold  0.5
nmsThreshold = 0.4  # Non-maximum suppression threshold  0.4
inpWidth = 416  # Width of network's input image
inpHeight = 416  # Height of network's input image

parser = argparse.ArgumentParser(
    description='Object Detection using YOLO in OPENCV')
parser.add_argument('--image', help='Path to image file.')
parser.add_argument('--point', default=None, help='Point')
parser.add_argument('--video', help='Path to video file.')
args = parser.parse_args()

# Load names of classes
# classesFile = "coco.names"
classesFile = "E:\zhangchunhui\VOT2019Challenge\SourceCode\TDE\Yolov3\coco.names"
classes = None
with open(classesFile, 'rt') as f:
    classes = f.read().rstrip('\n').split('\n')

colors = {}
np.random.seed(314)
for c in classes:
    color = np.random.random((3, )) * 0.6 + 0.4
    color = (color * 255).astype(np.int32).tolist()
    colors[c] = color

# Give the configuration and weight files for the model and load the network using them.
# modelConfiguration = "yolov3.cfg"
# modelWeights = "yolov3.weights"

modelConfiguration = "E:\zhangchunhui\VOT2019Challenge\SourceCode\TDE\Yolov3\yolov3.cfg"
modelWeights = "E:\zhangchunhui\VOT2019Challenge\SourceCode\TDE\Yolov3\yolov3.weights"

net = cv2.dnn.readNetFromDarknet(modelConfiguration, modelWeights)
net.setPreferableBackend(cv2.dnn.DNN_BACKEND_OPENCV)
net.setPreferableTarget(cv2.dnn.DNN_TARGET_CPU)


# Get the names of the output layers
def getOutputsNames(net):
    # Get the names of all the layers in the network
    layersNames = net.getLayerNames()
    # Get the names of the output layers, i.e. the layers with unconnected outputs
    return [layersNames[i[0] - 1] for i in net.getUnconnectedOutLayers()]


def draw_bboxes(image, bboxes, font_size=0.5, thresh=0.5, colors=None):
    """Draws bounding boxes on an image.

    Args:
        image: An image in OpenCV format
        bboxes: A dictionary representing bounding boxes of different object
            categories, where the keys are the names of the categories and the
            values are the bounding boxes. The bounding boxes of category should be
            stored in a 2D NumPy array, where each row is a bounding box (x1, y1,
            x2, y2, score).
        font_size: (Optional) Font size of the category names.
        thresh: (Optional) Only bounding boxes with scores above the threshold
            will be drawn.
        colors: (Optional) Color of bounding boxes for each category. If it is
            not provided, this function will use random color for each category.

    Returns:
        An image with bounding boxes.
    """

    # image = image.copy()
    for cat_name in bboxes:
        keep_inds = bboxes[cat_name][:, -1] > thresh
        cat_size = cv2.getTextSize(
            cat_name, cv2.FONT_HERSHEY_SIMPLEX, font_size, 2)[0]

        if colors is None:
            color = np.random.random((3, )) * 0.6 + 0.4
            color = (color * 255).astype(np.int32).tolist()
        else:
            color = colors[cat_name]

        for bbox in bboxes[cat_name][keep_inds]:
            bbox = bbox[0:4].astype(np.int32)
            if bbox[1] - cat_size[1] - 2 < 0:
                cv2.rectangle(image,
                              (bbox[0], bbox[1] + 2),
                              (bbox[0] + cat_size[0],
                               bbox[1] + cat_size[1] + 2),
                              color, -1
                              )
                cv2.putText(image, cat_name,
                            (bbox[0], bbox[1] + cat_size[1] + 2),
                            cv2.FONT_HERSHEY_SIMPLEX, font_size, (0, 0, 0), thickness=1
                            )
            else:
                cv2.rectangle(image,
                              (bbox[0], bbox[1] - cat_size[1] - 2),
                              (bbox[0] + cat_size[0], bbox[1] - 2),
                              color, -1
                              )
                cv2.putText(image, cat_name,
                            (bbox[0], bbox[1] - 2),
                            cv2.FONT_HERSHEY_SIMPLEX, font_size, (0, 0, 0), thickness=1
                            )
            cv2.rectangle(image,
                          (bbox[0], bbox[1]),
                          (bbox[2], bbox[3]),
                          color, 2
                          )
    return image


# Remove the bounding boxes with low confidence using non-maxima suppression
def postprocess(frame, outs):
    frameHeight = frame.shape[0]
    frameWidth = frame.shape[1]

    # Scan through all the bounding boxes output from the network and keep only the
    # ones with high confidence scores. Assign the box's class label as the class with the highest score.
    classIds = []
    confidences = []
    boxes = []
    # fout_txt = open('OUT.txt', 'w')
    fout_txt = open('E:\zhangchunhui\VOT2019Challenge\SourceCode\TDE\Yolov3\OUT.txt', 'w')
    for out in outs:
        for detection in out:
            scores = detection[5:]
            classId = np.argmax(scores)
            confidence = scores[classId]
            if confidence > confThreshold:
                center_x = int(detection[0] * frameWidth)
                center_y = int(detection[1] * frameHeight)
                width = int(detection[2] * frameWidth)
                height = int(detection[3] * frameHeight)
                left = int(center_x - width / 2)
                top = int(center_y - height / 2)
                classIds.append(classId)
                confidences.append(float(confidence))
                boxes.append([left, top, width, height])
                print(left, top, width, height, confidence, center_x, center_y, file = fout_txt)
    fout_txt.close()

    indices = cv2.dnn.NMSBoxes(boxes, confidences, confThreshold, nmsThreshold)
    indices = [i[0] for i in indices]
    classIds = np.array(classIds)[indices]
    confidences = np.array(confidences)[indices]
    boxes = np.array(boxes)[indices]

    bboxes = {}
    for cid, conf, box in zip(classIds, confidences, boxes):
        if classes[cid] not in bboxes.keys():
            bboxes[classes[cid]] = [
                [box[0], box[1], box[0]+box[2], box[1]+box[3], conf]]
        else:
            bboxes[classes[cid]].append(
                [box[0], box[1], box[0]+box[2], box[1]+box[3], conf])

    for k, v in bboxes.items():
        boxes = np.array(v)
        bboxes[k] = boxes
    draw_bboxes(frame, bboxes, thresh=confThreshold, colors=colors)


# Open the image file
if not os.path.isfile(args.image):
    print("Input image file ", args.image, " doesn't exist")
    sys.exit(1)
outputFile = args.image[:-4]+'_yolo_out_py.jpg'

frame = cv2.imread(args.image)

if args.point is not None:
    x, y, w, h = map(int, args.point.split(','))
    # x, y, w, h = map(int, args.point.split(' '))
    frame = frame[y:y+h, x:x+w, :]
    assert frame is not None

blob = cv2.dnn.blobFromImage(
    frame, 1/255, (inpWidth, inpHeight), [0, 0, 0], 1, crop=False)

net.setInput(blob)
outs = net.forward(getOutputsNames(net))
postprocess(frame, outs)

t, _ = net.getPerfProfile()
label = 'Inference time: %.2f ms' % (t * 1000.0 / cv2.getTickFrequency())
cv2.putText(frame, label, (0, 15),
            cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 255))

# Write the frame with the detection boxes
# cv2.imwrite(outputFile, frame.astype(np.uint8))
# cv2.imshow('demo', frame)
# cv2.waitKey()
